from pathlib import Path
from os import walk

import nornir
from nornir.core.task import Task
from rich.progress import Progress
from arista_lab.console import _print_failed_tasks

from nornir_napalm.plugins.tasks import napalm_cli, napalm_configure, napalm_get, napalm_confirm_commit  # type: ignore[import-untyped]
from nornir_jinja2.plugins.tasks import template_file  # type: ignore[import-untyped]


def _safe_push(task: Task, bar: Progress, *, config: str, title: str, replace: bool = False):
    r = task.run(
        task=napalm_configure,
        dry_run=False,
        replace=replace,
        configuration=config,
        revert_in=30,
    )
    if r.changed:
        bar.console.log(f"{task.host}: {title}\n\t{r.diff.replace('\n', '\n\t')}")
    r = task.run(task=napalm_confirm_commit)
    if r.changed:
        bar.console.log(f"{task.host}: {title}: {r.result}")


#############
# Templates #
#############


def apply_templates(
    nornir: nornir.core.Nornir,
    folder: Path,
    replace: bool = False,
    groups: bool = False,
) -> None:
    if not folder.exists():
        raise Exception(f"Could not find template folder {folder}")
    templates = []
    for dirpath, _, filenames in walk(folder):
        group = None
        if groups and len(dirpath.split("/")) > 1:
            # This refers to a group
            group = dirpath.split("/")[1]
        for file in filenames:
            if file.endswith(".j2"):
                templates.append((dirpath, file, group))
    with Progress() as bar:
        task_id = bar.add_task(
            "Apply configuration templates to devices",
            total=len(nornir.inventory.hosts) * len(templates),
        )

        def apply_templates(task: Task):
            for t in templates:
                if groups and not (
                    (group := t[2]) is None or group in task.host.groups
                ):
                    # Only apply templates specific to a group or templates with no group
                    bar.update(task_id, advance=1)
                    continue
                output = task.run(
                    task=template_file,
                    template=(template := t[1]),
                    path=t[0],
                    hosts=nornir.inventory.hosts,
                    groups=nornir.inventory.groups,
                )
                _safe_push(task, bar, config=output.result, title=template, replace=replace)
                bar.update(task_id, advance=1)


        results = nornir.run(task=apply_templates)
        if results.failed:
            _print_failed_tasks(bar, results)


###################
# Backup to flash #
###################


DIR_FLASH_CMD = "dir flash:"
BACKUP_FILENAME = "rollback-config"


def create_backups(nornir: nornir.core.Nornir) -> None:
    with Progress() as bar:
        task_id = bar.add_task(
            "Backup configuration to flash", total=len(nornir.inventory.hosts)
        )

        def create_backup(task: Task):
            r = task.run(task=napalm_cli, commands=[DIR_FLASH_CMD])
            for res in r:
                if BACKUP_FILENAME in res.result[DIR_FLASH_CMD]:
                    bar.console.log(f"{task.host}: Backup already present.")
                    bar.update(task_id, advance=1)
                    return
            task.run(
                task=napalm_cli,
                commands=[f"copy running-config flash:{BACKUP_FILENAME}"],
            )
            bar.console.log(f"{task.host}: Backup created.")
            bar.update(task_id, advance=1)

        results = nornir.run(task=create_backup)
        if results.failed:
            _print_failed_tasks(bar, results)


def restore_backups(nornir: nornir.core.Nornir) -> None:
    with Progress() as bar:
        task_id = bar.add_task(
            "Restore backup configuration from flash", total=len(nornir.inventory.hosts)
        )

        def restore_backup(task: Task):
            r = task.run(task=napalm_cli, commands=[DIR_FLASH_CMD])
            for res in r:
                if BACKUP_FILENAME in res.result[DIR_FLASH_CMD]:
                    task.run(
                        task=napalm_cli,
                        commands=[f"configure replace flash:{BACKUP_FILENAME}"],
                    )
                    # Intentionally not copying running-config to startup-config here.
                    # If there is a napalm_configure following a restore, configuration will be saved.
                    # This behaviour is acceptable, user can retrieve previous configuration in startup-config
                    # in case of mis-restoring the configuration.
                    bar.console.log(f"{task.host}: Backup restored.")
                    bar.update(task_id, advance=1)
                    return
            raise Exception(f"{task.host}: Backup not found.")

        results = nornir.run(task=restore_backup)
        if results.failed:
            _print_failed_tasks(bar, results)


def delete_backups(nornir: nornir.core.Nornir) -> None:
    with Progress() as bar:
        task_id = bar.add_task(
            "Delete backup on flash", total=len(nornir.inventory.hosts)
        )

        def delete_backup(task: Task):
            r = task.run(task=napalm_cli, commands=[DIR_FLASH_CMD])
            for res in r:
                if BACKUP_FILENAME in res.result[DIR_FLASH_CMD]:
                    task.run(
                        task=napalm_cli, commands=[f"delete flash:{BACKUP_FILENAME}"]
                    )
                    bar.console.log(f"{task.host}: Backup deleted.")
                    bar.update(task_id, advance=1)
                    return
            bar.console.log(f"{task.host}: Backup not found.")
            bar.update(task_id, advance=1)

        results = nornir.run(task=delete_backup)
        if results.failed:
            _print_failed_tasks(bar, results)


###############################
# Save and load configuration #
###############################


def save(nornir: nornir.core.Nornir, folder: Path) -> None:
    with Progress() as bar:
        task_id = bar.add_task(
            "Save lab configuration", total=len(nornir.inventory.hosts)
        )

        def save_config(task: Task):
            task.run(task=napalm_cli, commands=["copy running-config startup-config"])
            r = task.run(task=napalm_get, getters=["config"])
            config = folder / f"{task.host}.cfg"
            folder.mkdir(parents=True, exist_ok=True)
            with open(config, "w") as file:
                file.write(r[0].result["config"]["running"])
            bar.console.log(f"{task.host}: Configuration saved to {config}")
            bar.update(task_id, advance=1)

        results = nornir.run(task=save_config)
        if results.failed:
            _print_failed_tasks(bar, results)


def load(nornir: nornir.core.Nornir, folder: Path, replace: bool = False) -> None:
    with Progress() as bar:
        task_id = bar.add_task(
            "Load lab configuration", total=len(nornir.inventory.hosts)
        )

        def load_config(task: Task):
            config = folder / f"{task.host}.cfg"
            if not config.exists():
                raise Exception(
                    f"Configuration of {task.host} not found in folder {folder}"
                )
            output = task.run(
                task=template_file, template=f"{task.host}.cfg", path=folder
            )
            task.run(
                task=napalm_configure,
                dry_run=False,
                replace=replace,
                configuration=output.result,
                revert_in=30,
            )
            r = task.run(task=napalm_confirm_commit)
            bar.console.log(f"{task.host}: {r.result}")
            bar.update(task_id, advance=1)

        results = nornir.run(task=load_config)
        if results.failed:
            _print_failed_tasks(bar, results)



from pathlib import Path
from importlib.resources import files
from arista_lab import templates
from yaml import safe_load
import ipaddress

from typing import Any
import nornir
from nornir.core.task import Task
from rich.progress import Progress
from arista_lab.console import _print_failed_tasks

from nornir_jinja2.plugins.tasks import template_file  # type: ignore[import-untyped]

from . import _safe_push

def configure(nornir: nornir.core.Nornir, file: Path) -> None:
    DESCRIPTION_KEY = "description"
    IPV4_KEY = "ipv4"
    IPV4_SUBNET_KEY = "ipv4_subnet"
    IPV6_KEY = "ipv6"
    IPV6_SUBNET_KEY = "ipv6_subnet"
    ISIS_KEY = "isis"

    def _parse_links(file: Path):
        interfaces: dict[str, Any] = {}
        with file.open(mode="r", encoding="UTF-8") as f:
            links = safe_load(f)["links"]
            for link in links:
                if len(link["endpoints"]) != 2:
                    raise Exception(
                        f"Cannot parse '{file}': entry with 'endpoints' key must have a value in the format '['device1:etN', 'device2:etN']'"
                    )
                # for device_id, neighbor_id in (range(2), range(1,-1,-1)):
                device = link["endpoints"][0].split(":")[0]
                neighbor = link["endpoints"][1].split(":")[0]
                interface = link["endpoints"][0].split(":")[1]
                neighbor_interface = link["endpoints"][1].split(":")[1]
                if device not in interfaces:
                    interfaces[device] = {}
                if neighbor not in interfaces:
                    interfaces[neighbor] = {}
                interfaces[device][interface] = {
                    DESCRIPTION_KEY: f"to {neighbor} {neighbor_interface}"
                }
                interfaces[neighbor][neighbor_interface] = {
                    DESCRIPTION_KEY: f"to {device} {interface}"
                }
                if ISIS_KEY in link:
                    interfaces[device][interface].update({ISIS_KEY: link[ISIS_KEY]})
                    interfaces[neighbor][neighbor_interface].update(
                        {ISIS_KEY: link[ISIS_KEY]}
                    )
                if IPV4_SUBNET_KEY in link:
                    network = ipaddress.ip_network(link[IPV4_SUBNET_KEY])
                    if network.prefixlen != 31:
                        raise Exception(f"Subnet {network} is not a /31 subnet")
                    interfaces[device][interface].update(
                        {IPV4_KEY: f"{network[0]}/{network.prefixlen}"}
                    )
                    interfaces[neighbor][neighbor_interface].update(
                        {IPV4_KEY: f"{network[1]}/{network.prefixlen}"}
                    )
                if IPV6_SUBNET_KEY in link:
                    network = ipaddress.ip_network(link[IPV6_SUBNET_KEY])
                    if network.prefixlen != 127:
                        raise Exception(f"Subnet {network} is not a /127 subnet")
                    interfaces[device][interface].update(
                        {IPV6_KEY: f"{network[0]}/{network.prefixlen}"}
                    )
                    interfaces[neighbor][neighbor_interface].update(
                        {IPV6_KEY: f"{network[1]}/{network.prefixlen}"}
                    )
        return interfaces

    links = _parse_links(file)
    with Progress() as bar:
        task_id = bar.add_task(
            "Configure point-to-point interfaces", total=len(nornir.inventory.hosts)
        )

        def configure_interfaces(task: Task):
            for interface, params in links[task.host.name].items():
                p = files(templates) / "interfaces"
                output = task.run(
                    task=template_file,
                    template="point-to-point.j2",
                    path=p,
                    interface={"name": interface, **params},
                )
                _safe_push(task, bar, config=output.result, title=f"Interface {interface} ({'IPv4' if IPV4_KEY in params else ''} {'IPv6' if IPV6_KEY in params else ''} {'ISIS' if ISIS_KEY in params else ''}): {params[DESCRIPTION_KEY]}")
                bar.update(task_id, advance=1)

        results = nornir.run(task=configure_interfaces)
        if results.failed:
            _print_failed_tasks(bar, results)



from importlib.resources import files
from arista_lab import templates
from datetime import datetime, timedelta
import ipaddress

import requests
import nornir
from nornir.core.task import Task
from nornir.core.filter import F
from rich.progress import Progress
from arista_lab.console import _print_failed_tasks

from nornir_jinja2.plugins.tasks import template_file  # type: ignore[import-untyped]

from . import _safe_push


def configure(nornir: nornir.core.Nornir, group: str, neighbor_group: str) -> None:
    def _build_vars(asn: int):
        start_time = datetime.now() - timedelta(days=10)
        url = f"https://stat.ripe.net/data/announced-prefixes/data.json?resource=AS{asn}&starttime={start_time.strftime('%Y-%m-%dT%H:%M')}"
        r = requests.get(url)
        if r.ok:
            prefixes = []
            for prefix in r.json()["data"]["prefixes"]:
                prefixes.append(prefix["prefix"])
        else:
            raise Exception(f"Could not get announced prefixes for AS{asn}")
        networks = []
        for network in [ipaddress.ip_network(p) for p in prefixes]:
            if not any(
                (network != n and network.overlaps(n))
                for n in [ipaddress.ip_network(p) for p in prefixes]
            ):
                networks.append(network)

        hosts = []
        hosts_ipv6 = []
        prefixes = []
        prefixes_ipv6 = []
        for network in networks:
            if network.version == 4:
                hosts.append(f"{next(network.hosts())}/{network.prefixlen}")
                prefixes.append(str(network))
            elif network.version == 6:
                hosts_ipv6.append(f"{next(network.hosts())}/{network.prefixlen}")
                prefixes_ipv6.append(str(network))

        return {
            "hosts": hosts,
            "hosts_ipv6": hosts_ipv6,
            "prefixes": prefixes,
            "prefixes_ipv6": prefixes_ipv6,
        }

    with Progress() as bar:
        task_id = bar.add_task(
            "Configure peering devices",
            total=len(nornir.inventory.children_of_group(group)),
        )

        def configure_peering(task: Task):
            MAX_LOOPBACKS = 2100
            vars = _build_vars(task.host.data["asn"])
            bar.console.log(
                f"{task.host}: Configuring {len(vars['prefixes'])} IPv4 prefixes for ISP {task.host.data['isp']}"
            )
            # bar.console.log(f"{task.host}: {vars['prefixes']}")
            bar.console.log(
                f"{task.host}: Configuring {len(vars['prefixes_ipv6'])} IPv6 prefixes for ISP {task.host.data['isp']}"
            )
            # bar.console.log(f"{task.host}: {vars['prefixes_ipv6']}")
            vars.update(
                {
                    "name": task.host.data["isp"],
                    "asn": task.host.data["asn"],
                    "description": task.host.data["description"],
                    "as_path_length": task.host.data["as_path_length"],
                    "max_loopback": MAX_LOOPBACKS,
                    "neighbor_name": task.nornir.inventory.groups[neighbor_group].data[
                        "network_name"
                    ],
                    "neighbor_ipv4": task.host.data["neighbor_ipv4"],
                    "neighbor_ipv6": task.host.data["neighbor_ipv6"],
                    "neighbor_as": task.nornir.inventory.groups[neighbor_group].data[
                        "asn"
                    ],
                }
            )

            p = files(templates) / "peering"
            output = task.run(task=template_file, template="isp.j2", path=p, vars=vars)
            _safe_push(task, bar, config=output.result, title=f"Peering with {task.nornir.inventory.groups[neighbor_group].data['network_name']}")
            bar.update(task_id, advance=1)

        results = nornir.filter(F(groups__contains=group)).run(task=configure_peering)
        if results.failed:
            _print_failed_tasks(bar, results)



from pathlib import Path
import pickle
import snappi # type: ignore[import-untyped]
import snappi_ixnetwork # type: ignore[import-untyped]
import logging
from rich.table import Table
from rich.console import Console
import urllib3

urllib3.disable_warnings()
console = Console()
logger = logging.getLogger(__name__)

snappi_ixnetwork_session_file = Path("./.snappi-api-session")

def configure(
    api: snappi.Api,
    config: snappi.Config,
) -> None:
    """Configure the flows on the traffic generator"""
    try:
        api.set_config(config)
    except Exception as e:
        logger.error(e) # snappi_ixnetwork.exceptions.IxNetworkException is raised here
    if isinstance(api, snappi_ixnetwork.Api):
        with snappi_ixnetwork_session_file.open(mode="wb") as fd:
            pickle.dump(api.get_config(), fd)

def _get_traffic_stats(api: snappi.Api) -> tuple[list, list]:
    request = api.metrics_request()
    request.choice = request.PORT
    port_stats = api.get_metrics(request).port_metrics
    if port_stats is None:
        port_stats = []

    request = api.metrics_request()
    request.choice = request.FLOW
    flow_stats = api.get_metrics(request).flow_metrics
    if flow_stats is None:
        flow_stats = []
    return port_stats, flow_stats

def _print_traffic_stats(
    port_stats=None,
    flow_stats=None,
    bgpv4_stats=None,
    bgpv6_stats=None,
) -> None:
    if port_stats is not None:
        table = Table(title="Port Metrics")
        table.add_column("Port")
        table.add_column("Tx Frames")
        table.add_column("Tx Bytes")
        table.add_column("Rx Frames")
        table.add_column("Rx Bytes")
        table.add_column("Tx FPS")
        table.add_column("Rx FPS")
        for stat in port_stats:
            table.add_row(
                stat.name,
                str(stat.frames_tx),
                str(stat.bytes_tx),
                str(stat.frames_rx),
                str(stat.bytes_rx),
                str(stat.frames_tx_rate),
                str(stat.frames_rx_rate),
            )
        console.print(table)
    if flow_stats is not None:
        table = Table(title="Flow Metrics")
        table.add_column("Flow")
        table.add_column("Rx Frames")
        table.add_column("Rx Bytes")
        table.add_column("Transmit State")
        for stat in flow_stats:
            table.add_row(
                stat.name,
                str(stat.frames_rx),
                str(stat.bytes_rx),
                stat.transmit,
            )
        console.print(table)
    if bgpv4_stats is not None:
        table = Table(title="BGPv4 Metrics")
        table.add_column("Name")
        table.add_column("Session State")
        table.add_column("Session Flaps")
        table.add_column("Routes Advertised")
        table.add_column("Routes Received")
        table.add_column("Route Withdraws Tx")
        table.add_column("Route Withdraws Rx")
        table.add_column("Keepalives Tx")
        table.add_column("Keepalives Rx")
        for stat in bgpv4_stats:
            table.add_row(
                stat.name,
                stat.session_state,
                str(stat.session_flap_count),
                str(stat.routes_advertised),
                str(stat.routes_received),
                str(stat.route_withdraws_sent),
                str(stat.route_withdraws_received),
                str(stat.keepalives_sent),
                str(stat.keepalives_received),
            )
        console.print(table)
    if bgpv6_stats is not None:
        table = Table(title="BGPv6 Metrics")
        table.add_column("Name")
        table.add_column("Session State")
        table.add_column("Session Flaps")
        table.add_column("Routes Advertised")
        table.add_column("Routes Received")
        table.add_column("Route Withdraws Tx")
        table.add_column("Route Withdraws Rx")
        table.add_column("Keepalives Tx")
        table.add_column("Keepalives Rx")
        for stat in bgpv6_stats:
            table.add_row(
                stat.name,
                stat.session_state,
                str(stat.session_flap_count),
                str(stat.routes_advertised),
                str(stat.routes_received),
                str(stat.route_withdraws_sent),
                str(stat.route_withdraws_received),
                str(stat.keepalives_sent),
                str(stat.keepalives_received),
            )
        console.print(table)

def start(api: snappi.Api) -> None:
    """Start the flows on the traffic generator"""
    control_state = api.control_state()
    control_state.choice = control_state.TRAFFIC
    control_state.traffic.choice = control_state.traffic.FLOW_TRANSMIT
    control_state.traffic.flow_transmit.state = control_state.traffic.flow_transmit.START
    try:
        res = api.set_control_state(control_state)
        for warning in res.warnings:
            logger.warning(warning)
    except Exception as e:
        logger.error(e)

def stop(api: snappi.Api) -> None:
    """Stop the flows on the traffic generator"""
    control_state = api.control_state()
    control_state.choice = control_state.TRAFFIC
    control_state.traffic.choice = control_state.traffic.FLOW_TRANSMIT
    control_state.traffic.flow_transmit.state = (
        control_state.traffic.flow_transmit.STOP
    )
    try:
        res = api.set_control_state(control_state)
        for warning in res.warnings:
            logger.warning(warning)
    except Exception as e:
        logger.error(e)

def stats(api: snappi.Api) -> None:
    port_stats, flow_stats = _get_traffic_stats(api)
    _print_traffic_stats(port_stats=port_stats, flow_stats=flow_stats)



__version__ = "0.1.0"

import ssl
from pyeapi.eapilib import HttpsEapiConnection  # type: ignore[import-untyped]


class LabEapiConnection(HttpsEapiConnection):
    """
    https://arista.my.site.com/AristaCommunity/s/article/Python-3-10-and-SSLV3-ALERT-HANDSHAKE-FAILURE-error
    """

    def __init__(self, **kwargs):
        # Use TLSv1.2
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.minimum_version = ssl.TLSVersion.TLSv1_2
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE

        # Using the EOS default ciphers
        context.set_ciphers(
            "AES256-SHA:DHE-RSA-AES256-SHA:AES128-SHA:DHE-RSA-AES128-SHA"
        )

        if "context" in kwargs:
            del kwargs["context"]

        super(LabEapiConnection, self).__init__(**kwargs, context=context)



#!/usr/bin/env python
from enum import Enum
import pickle
from typing import Literal
import nornir
import click
import sys

import logging
from rich.console import Console
from pathlib import Path
from nornir.core.filter import F
from rich.logging import RichHandler

import arista_lab.config
import arista_lab.traffic
import arista_lab.config.interfaces
import arista_lab.config.peering
import snappi # type: ignore[import-untyped]

console = Console()

logger = logging.getLogger(__name__)

class Log(str, Enum):
    """Represent log levels from logging module as immutable strings."""

    CRITICAL = logging.getLevelName(logging.CRITICAL)
    ERROR = logging.getLevelName(logging.ERROR)
    WARNING = logging.getLevelName(logging.WARNING)
    INFO = logging.getLevelName(logging.INFO)
    DEBUG = logging.getLevelName(logging.DEBUG)


LogLevel = Literal[Log.CRITICAL, Log.ERROR, Log.WARNING, Log.INFO, Log.DEBUG]

def setup_logging(level: LogLevel = Log.INFO, file: Path | None = None) -> None:
    """Configure logging for Python.

    If a file is provided, logs will also be sent to the file in addition to stdout.
    If a file is provided and logging level is DEBUG, only the logging level INFO and higher will
    be logged to stdout while all levels will be logged in the file.

    Args:
    ----
        level: Python logging level
        file: Send logs to a file

    """
    # Get loggers
    loggers = ["arista_lab", "snappi", "snappi_ixnetwork", "ixnetwork_restpy.connection", "pyeapi"]
    loglevel = getattr(logging, level.upper())
    for logger_name in loggers:
        logging.getLogger(logger_name).setLevel(loglevel)
    # Silence the logging of chatty Python modules when level is INFO
    if loglevel == logging.INFO:
        logging.getLogger("pyeapi").setLevel(logging.CRITICAL)
    # Add RichHandler for stdout
    rich_handler = RichHandler(rich_tracebacks=True, tracebacks_show_locals=False)
    # Show Python module in stdout at DEBUG level
    fmt_string = (
        "[%(name)s] %(message)s"
        if loglevel == logging.DEBUG
        else "%(message)s"
    )
    formatter = logging.Formatter(fmt=fmt_string, datefmt="[%X]")
    rich_handler.setFormatter(formatter)
    for logger_name in loggers:
        logging.getLogger(logger_name).addHandler(rich_handler)
    # Add FileHandler if file is provided
    if file:
        file_handler = logging.FileHandler(file)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        file_handler.setFormatter(formatter)
        for logger_name in loggers:
            logging.getLogger(logger_name).addHandler(file_handler)
        # If level is DEBUG and file is provided, do not send DEBUG level to stdout
        if loglevel == logging.DEBUG:
            rich_handler.setLevel(logging.INFO)

def _init_nornir(ctx: click.Context, param, value: Path) -> nornir.core.Nornir:
    try:
        return nornir.InitNornir(config_file=str(value), core={"raise_on_error": False})
    except Exception as exc:
        ctx.fail(f"Unable to initialize Nornir with config file '{value}': {str(exc)}")


def _read_otg_config(ctx: click.Context, param, value: Path) -> snappi.Config:
    try:
        config = ctx.obj["snappi_api"].config()
        with value.open(mode="r", encoding="UTF-8") as fd:
            config.deserialize(fd.read())
        logger.debug(f"OTG Configuration: {config}")
        return config
    except Exception as exc:
        ctx.fail(f"Unable to read OTG Configuration from file '{value}': {str(exc)}")

@click.group()
@click.option(
    "-n",
    "--nornir",
    "nornir",
    default="nornir.yaml",
    type=click.Path(exists=True, readable=True, dir_okay=False, path_type=Path),
    callback=_init_nornir,
    show_default=True,
    show_envvar=True,
    help="Nornir configuration in YAML format.",
)
@click.option(
    "--log-file",
    help="Send the logs to a file. If logging level is DEBUG, only INFO or higher will be sent to stdout.",
    show_envvar=True,
    type=click.Path(file_okay=True, dir_okay=False, writable=True, path_type=Path),
)
@click.option(
    "--log-level",
    "-l",
    help="Python logging level",
    default=logging.getLevelName(logging.INFO),
    show_envvar=True,
    show_default=True,
    type=click.Choice(
        [Log.CRITICAL, Log.ERROR, Log.WARNING, Log.INFO, Log.DEBUG],
        case_sensitive=False,
    ),
)
@click.pass_context
def cli(ctx, nornir: nornir.core.Nornir, log_level: LogLevel, log_file: Path) -> None:
    ctx.ensure_object(dict)
    ctx.obj["nornir"] = nornir
    setup_logging(log_level, log_file)

@cli.command(help="Create or delete device configuration backups to flash")
@click.pass_obj
@click.option(
    "--delete/--no-delete",
    default=False,
    help="Delete the backup on the device flash",
    show_default=True,
)
def backup(obj: dict, delete: bool) -> None:
    if delete:
        arista_lab.config.delete_backups(obj["nornir"])
    else:
        arista_lab.config.create_backups(obj["nornir"])


@cli.command(help="Restore configuration backups from flash")
@click.pass_obj
def restore(obj: dict) -> None:
    arista_lab.config.restore_backups(obj["nornir"])

@cli.command(help="Save configuration to a folder")
@click.pass_obj
@click.option(
    "--folder",
    "folder",
    type=click.Path(writable=True, path_type=Path),
    required=True,
    help="Configuration backup folder",
)
def save(obj: dict, folder: Path) -> None:
    arista_lab.config.save(obj["nornir"], folder)

@cli.command(help="Load configuration from a folder")
@click.pass_obj
@click.option(
    "--folder",
    "folder",
    type=click.Path(writable=True, path_type=Path),
    required=True,
    help="Configuration backup folder",
)
@click.option(
    "--replace/--merge",
    default=False,
    show_default=True,
    help="Replace or merge the configuration on the device",
)
def load(obj: dict, folder: Path, replace: bool) -> None:
    arista_lab.config.create_backups(obj["nornir"])
    arista_lab.config.load(obj["nornir"], folder, replace=replace)

@cli.command(help="Apply configuration templates")
@click.pass_obj
@click.option(
    "--folder",
    "folder",
    type=click.Path(writable=True, path_type=Path),
    required=True,
    help="Configuration template folder",
)
@click.option(
    "--groups/--no-groups",
    default=False,
    show_default=True,
    help="The template folder contains subfolders with Nornir group names",
)
@click.option(
    "--replace/--merge",
    default=False,
    show_default=True,
    help="Replace or merge the configuration on the device",
)
def apply(obj: dict, folder: Path, groups: bool, replace: bool) -> None:
    arista_lab.config.create_backups(obj["nornir"])
    arista_lab.config.apply_templates(
        obj["nornir"], folder, replace=replace, groups=groups
    )

@cli.command(help="Configure point-to-point interfaces")
@click.pass_obj
@click.option(
    "--links",
    "links",
    type=click.Path(exists=True, readable=True, path_type=Path),
    required=True,
    help="YAML File describing lab links",
)
def interfaces(obj: dict, links: Path) -> None:
    arista_lab.config.create_backups(obj["nornir"])
    arista_lab.config.interfaces.configure(obj["nornir"], links)

@cli.command(help="Configure peering devices")
@click.pass_obj
@click.option(
    "--group", "group", type=str, required=True, help="Nornir group of peering devices"
)
@click.option(
    "--backbone",
    "backbone",
    type=str,
    required=True,
    help="Nornir group of the backbone",
)
def peering(obj: dict, group: str, backbone: str) -> None:
    arista_lab.config.create_backups(obj["nornir"].filter(F(groups__contains=group)))
    arista_lab.config.peering.configure(obj["nornir"], group, backbone)

@cli.group(help="Control traffic generator")
@click.option(
    "--otg-api",
    "otg_api",
    type=str,
    required=True,
    show_envvar=True,
    help="Open Traffic Generator server",
)
@click.option(
    "--extension",
    "snappi_extension",
    type=click.Choice(["ixnetwork"]),
    show_envvar=True,
    help="Snappi Extension to use. Supported extension is 'ixnetwork'.",
)
@click.pass_context
def traffic(
    ctx: click.Context, otg_api: str, snappi_extension: Literal["ixnetwork"] | None,
) -> None:
    try:
        ctx.obj["snappi_api"] = snappi.api(
                location=otg_api,
                ext=snappi_extension,
            )
    except Exception as e:
        logger.error(e)
        ctx.exit(1)
    if snappi_extension == "ixnetwork" and arista_lab.traffic.snappi_ixnetwork_session_file.exists():
        with arista_lab.traffic.snappi_ixnetwork_session_file.open(mode="rb") as fd:
            ctx.obj["snappi_api"]._config = pickle.load(fd)

@traffic.command(help="Configure traffic generator")
@click.argument(
    "config",
    default="otg.yaml",
    type=click.Path(exists=True, dir_okay=False, readable=True, path_type=Path),
    callback=_read_otg_config,
)
@click.pass_obj
def configure(
    obj: dict,
    config: snappi.Config,
) -> None:
    arista_lab.traffic.configure(api=obj["snappi_api"], config=config)

@traffic.command(help="Start the flows on the traffic generator")
@click.pass_obj
def start(
    obj: dict,
) -> None:
    arista_lab.traffic.start(api=obj["snappi_api"])

@traffic.command(help="Stop the flows on the traffic generator")
@click.pass_obj
def stop(
    obj: dict,
) -> None:
    arista_lab.traffic.stop(api=obj["snappi_api"])

@traffic.command(help="Get the flow statistics from the traffic generatorr")
@click.pass_obj
def stats(
    obj: dict,
) -> None:
    arista_lab.traffic.stats(api=obj["snappi_api"])

def main() -> None:
    try:
        sys.exit(cli(auto_envvar_prefix="LAB"))
    except Exception:
        console.print_exception()
        sys.exit(1)

if __name__ == '__main__':
    main()



from rich.progress import Progress
from nornir.core.exceptions import NornirSubTaskError
from nornir.core.task import AggregatedResult


def _print_failed_tasks(bar: Progress, results: AggregatedResult) -> None:
    for host, multi_results in results.items():
        for r in multi_results:
            if r.failed:
                if isinstance(r.exception, NornirSubTaskError):
                    # Do not display NornirSubTaskError
                    continue
                bar.console.log(
                    f"Task failed for device {host}: {r.exception.__class__.__name__}"
                )
                bar.console.log(r.exception)
